#include <VL53L1X.h>
#include <Wire.h>
#include <Servo.h>
#include <EEPROM.h>

#define TARGET_DISTANCE_EEPROM_ADDRESS 1

#define BUTTON_PIN 2
#define SERVO_PIN 9
#define LED_PIN 13 // built in led pin = 13

#define MIN_PULSE_WIDTH      544
#define LOWER_PULSE_WIDTH    1000 // lower callibration position
#define DEFAULT_PULSE_WIDTH  1500
#define UPPER_PULSE_WIDTH    2000 // upper callibration position
#define MAX_PULSE_WIDTH      2400

#define PULSE_WIDTH_REBOUND 5
#define PULSE_WIDTH_TOLERANCE 1 
#define PULSE_WIDTH_ADJUSTMENT_WEIGHT 0.10

#define MIN_MAX_ALERT_PULSE_WIDTH_TOLERANCE 10 

#define SENSOR_ANGLE_RAD 0.698131701 // sensor pointing 40 deg down

#define MAIN_LOOP_DELAY 100

#define DEFAULT_TARGET_DISTANCE 800
#define MAX_TARGET_DISTANCE 3000
#define MIN_TARGET_DISTANCE 500

#define MIN_MEASURED_DISTANCE 500

#define GEAR_RATIO 0.3

#define BLINK_DELAY 50UL
#define BLINK_PAUSE_DELAY_SHORT 100UL
#define BLINK_PAUSE_DELAY_MED 500UL
#define BLINK_PAUSE_DELAY_LONG 2000UL


Servo myservo;
VL53L1X sensor;

int measuredDistance = DEFAULT_TARGET_DISTANCE;

//int targetDistance = DEFAULT_TARGET_DISTANCE;
struct TargetDistances {
  int distanceCenter = DEFAULT_TARGET_DISTANCE; //target distance when centered
  int distanceLower = DEFAULT_TARGET_DISTANCE; // target distance in lower position (facing down)
  int distanceUpper = DEFAULT_TARGET_DISTANCE; // target distance in upper position (facing up)
};

struct TargetDistances distances;

//int heightCalculated = DEFAULT_TARGET_DISTANCE; // will be used for actual angle calculations when new measurement is taken

int currentPulseWidth = DEFAULT_PULSE_WIDTH;
//int targetDistancePulseWidth = DEFAULT_PULSE_WIDTH;
double servoAngleToPulseWidthRatio = (MAX_PULSE_WIDTH - MIN_PULSE_WIDTH) / (GEAR_RATIO * PI); // to convert angle adjustment to pulse width adjustment

double targetAngleRad = SENSOR_ANGLE_RAD; //real sensor-to-horizon angle in default position. assume this is close to servo middle position
double currentAngleRad = SENSOR_ANGLE_RAD; // this will be calculated based on distance measurements
//float smoothedAngleRad = SENSOR_ANGLE_RAD;

// status flags
bool LedBlinkState = false;
bool callibrating = false;
int callibrationStep = 0;
bool trackerEnabled = false;
bool maxPositionAlert = false;
bool sensorFailureAlert = false;

//LED timer
unsigned long ledBlinkTime = 0UL;
 
void setup() {
  pinMode(LED_PIN, OUTPUT);
  pinMode(BUTTON_PIN, INPUT);
  
  Wire.begin();
  Wire.setClock(400000); // use 400 kHz I2C
  sensor.setTimeout(500);
  if (!sensor.init()){ //sesor error
    while (1){
      toggleLed(); // blink fast if cannot initialise sensor
      delay(50);
    }
  }

  sensor.setDistanceMode(VL53L1X::Long); // Long = 4m range
  sensor.setMeasurementTimingBudget(50000);
  sensor.startContinuous(50);
  
  myservo.attach(SERVO_PIN);
  
  EEPROM.get(TARGET_DISTANCE_EEPROM_ADDRESS, distances); // read saved target distance from eeprom

  if(distances.distanceCenter <= 0){
    distances.distanceCenter = DEFAULT_TARGET_DISTANCE;
    distances.distanceLower = DEFAULT_TARGET_DISTANCE;
    distances.distanceUpper = DEFAULT_TARGET_DISTANCE;
  } else {
    distances.distanceCenter = max(min(distances.distanceCenter, MAX_TARGET_DISTANCE), MIN_TARGET_DISTANCE);
    // distance in upper position should be higher than center
    distances.distanceLower = max(min(distances.distanceLower, MAX_TARGET_DISTANCE), distances.distanceCenter);
    // distance in upper position should be lower than center
    distances.distanceUpper = max(min(distances.distanceUpper, distances.distanceCenter), MIN_TARGET_DISTANCE); 
  }

  // move servo to default position
  myservo.writeMicroseconds(currentPulseWidth);
  disableLed();
}

void loop() {
  handleButtonActions();
  handleLedActions();
  
  //measure distance and calculate servo adjustment
  if(trackerEnabled){

    int servoPulseWidthAdjustment = 0; 
    
    sensor.read();
    
    if(sensor.ranging_data.range_status == 0){ // 0 = RangeValid or 7 = WrapTargetFail
      measuredDistance = sensor.ranging_data.range_mm;
      sensorFailureAlert = false;

      int heightCalculated = calculateHeight();

      if(measuredDistance > heightCalculated){ // otherwise angle calculation fails

        currentAngleRad = calculateCurrentAngleRad(heightCalculated, measuredDistance); // calculate estimated current sensor-to-horizon angle
        
        float angleDiff = currentAngleRad - targetAngleRad; // positive value means sensor needs to be moved down
        servoPulseWidthAdjustment = int(angleDiff * servoAngleToPulseWidthRatio * PULSE_WIDTH_ADJUSTMENT_WEIGHT);

        if(abs(servoPulseWidthAdjustment) <= PULSE_WIDTH_TOLERANCE){ //avoid small adjustments
          servoPulseWidthAdjustment = 0;
        }

        currentPulseWidth = currentPulseWidth + servoPulseWidthAdjustment; // increasing this value moves sensor up (decreasing sensor-to-horizon angle)
        
      }  else if (currentPulseWidth < LOWER_PULSE_WIDTH){   
        currentPulseWidth = currentPulseWidth + PULSE_WIDTH_REBOUND; // rebound up if measured distance is too low most likely bacause wheel part 
      }
    } else {
      sensorFailureAlert = true;
  
    }
    
    currentPulseWidth = max(min(currentPulseWidth, MAX_PULSE_WIDTH), MIN_PULSE_WIDTH); // make sure it stays between min and max values

    int pulseWidthDistanceFromBoundary = min(MAX_PULSE_WIDTH - currentPulseWidth, currentPulseWidth - MIN_PULSE_WIDTH);
    if(pulseWidthDistanceFromBoundary <= MIN_MAX_ALERT_PULSE_WIDTH_TOLERANCE ){
      maxPositionAlert = true;
    } else {
      maxPositionAlert = false;
    }
  } else {
    // use default position if tracker disabled
    if(!callibrating){
      currentPulseWidth = DEFAULT_PULSE_WIDTH; 
    }
    maxPositionAlert = false;
    sensorFailureAlert = false;
  }

  // move servo
  myservo.writeMicroseconds(currentPulseWidth);
  
  delay(MAIN_LOOP_DELAY);
}


void handleLedActions(){ // blink led
  if(maxPositionAlert or sensorFailureAlert){
    enableLed();
  } else { 
    unsigned long timeDiff = millis() - ledBlinkTime;
    unsigned long blinkDelay = BLINK_DELAY;

    if(LedBlinkState){ // if already enabled
      blinkDelay = BLINK_DELAY;
    } else if(callibrating){
      blinkDelay = BLINK_PAUSE_DELAY_SHORT;
    } else if(trackerEnabled){
      blinkDelay = BLINK_PAUSE_DELAY_MED;
    } else {
      blinkDelay = BLINK_PAUSE_DELAY_LONG;
    }
 
    if(timeDiff > blinkDelay){
      toggleLed();
      ledBlinkTime = millis(); // reset timer
    }
  } 
}


void handleButtonActions(){
  if(digitalRead(BUTTON_PIN) == HIGH){   
    unsigned long buttonPressedTime = millis();
    while(digitalRead(BUTTON_PIN) == HIGH){ // stop everything else while pressed
      enableLed();
      delay(10);
    }
    //button release 
    disableLed();
    unsigned long timeDiff = millis() - buttonPressedTime;

    if(timeDiff > 3000UL){ 
      // 3 seconds
      callibrating = !callibrating; //toggle callibrating status
      if(callibrating){ // disable tracker
        trackerEnabled = false;
        callibrationStep = 1; 
        //confirm callibration start
        delay(100); 
        myservo.writeMicroseconds(UPPER_PULSE_WIDTH);
        delay(500); 
        myservo.writeMicroseconds(LOWER_PULSE_WIDTH);
        delay(500); 
        myservo.writeMicroseconds(DEFAULT_PULSE_WIDTH);
        delay(500); 
      } else {
        callibrationStep = 0; 
      }

    } else if(timeDiff > 1000UL){  
      // 1 second
      if(callibrating){ // read and set target distances
        if(callibrationStep == 1){
          // measure center distance
          distances.distanceCenter = measureDistance();
          delay(100);
          // move to upper position
          currentPulseWidth = UPPER_PULSE_WIDTH;
          callibrationStep = callibrationStep + 1;

        } else if(callibrationStep == 2){
          // measure upper distance
          distances.distanceUpper = measureDistance();
          delay(100);
          // move to lower position
          currentPulseWidth = LOWER_PULSE_WIDTH;
          callibrationStep = callibrationStep + 1;

        } else if(callibrationStep == 3){
          // measure lower distance
          distances.distanceLower = measureDistance();
          //save all distancesto eeprom
          EEPROM.put(TARGET_DISTANCE_EEPROM_ADDRESS, distances);
          callibrationStep = 0; 
          callibrating = false;
          trackerEnabled = true;

          //confirm callibration end
          delay(100); 
          myservo.writeMicroseconds(UPPER_PULSE_WIDTH);
          delay(500); 
          myservo.writeMicroseconds(LOWER_PULSE_WIDTH);
          delay(500); 
          myservo.writeMicroseconds(DEFAULT_PULSE_WIDTH);
          delay(500); 
        } 
      }    

    } else if(timeDiff > 10UL){ 
      // short
      if(!callibrating){
        trackerEnabled = !trackerEnabled; // toggle tracker state
      }
    }
  }
}


int measureDistance(){
  sensor.read();
  delay(100);
  while(sensor.ranging_data.range_status != 0){ //ensure correct reading
    sensor.read();
    delay(100);
  }
  return sensor.ranging_data.range_mm;
}


int calculateHeight(){
  int targetDistance = interpolateTargetDistance();
  return int(double(targetDistance) * sin(targetAngleRad)); // this height will be used to calculate angle based on distance measurements
}


int interpolateTargetDistance(){
  if(currentPulseWidth < DEFAULT_PULSE_WIDTH){
    return map(currentPulseWidth, LOWER_PULSE_WIDTH, DEFAULT_PULSE_WIDTH, distances.distanceLower, distances.distanceCenter);
  } else {
    return map(currentPulseWidth, DEFAULT_PULSE_WIDTH, UPPER_PULSE_WIDTH, distances.distanceCenter, distances.distanceUpper);
  }
}


double calculateCurrentAngleRad(int height, int distance){
  return asin(double(height) / double(distance));
}


void enableLed(){
  if(LedBlinkState == false){
    digitalWrite(LED_PIN, HIGH);
    LedBlinkState = true;
  }
}


void disableLed(){
  if(LedBlinkState == true){
    digitalWrite(LED_PIN, LOW);
    LedBlinkState = false;
  }
}


void toggleLed(){
  if(LedBlinkState){ // if enabled
    disableLed();
  } else {
    enableLed();
  }
}
